
%if 0

RxANSI: Combined ANSI and LESS driver. AMIS v3.6 compliant.
Public Domain by Henrik Haftmann and Christian Masloch

RxANSIT.ASM - Transient code of RxANSI; command line parser and setup

Read RxANSI.DOC for more information.

Assemble: NASM.EXE -fbin -oRXANSI.COM RXANSIT.ASM

Options to append to that line:
-D_I2F=0	Don't hook interrupt 2Fh
-D_LESS=0	Don't parse LESS
-D_PSPRELOC=0	Don't relocate PSP
These options also work with the supplied make.bat script.

(German translation follows)
Assemblieren: NASM.EXE -fbin -oRXANSI.COM RXANSIT.ASM

Optionen, die hinter dieser Zeile stehen:
-D_I2F=0	Interrupt 2Fh nicht umleiten
-D_LESS=0	LESS nicht interpretieren
-D_PSPRELOC=0	PSP nicht relozieren
-D_LANG="DEU"	Deutsche Texte nutzen
Diese Optionen funktionieren auch mit dem bereitgestellten Skript make.bat.

%endif

	%assign _186 0
	%assign _386 0

%include "CMMACROS.MAC"
%include "AMIS.MAC"

	verdef RXANSI,C, 1,10		; Version number of last compatible
	verdef RXANSI,P, 1,10		; Private version

	numdef I2F, 1			; hook interrupt 2Fh
	numdef LESS, 1			; parse LESS sequences
	numdef PSPRELOC, 1		; relocate PSP
	numdef UPXPAD, 0		; pad executable to >=3062 bytes for UPX
	strdef LANG, "ENG"

		; Bits of the option selector (hold in bp)
_holdenvironment	equ 0001h
_installlow		equ 0002h
_installnew		equ 0004h
_uninstall		equ 0008h
_lessstate		equ 0040h		; must match LESS bit in ctrl0
_ansistate		equ 0080h		; must match ANSI bit in ctrl0
_state			equ _ansistate|_lessstate
_lessset		equ _lessstate<<2
_ansiset		equ _ansistate<<2
_foundresident		equ 1000h
_incompatresident	equ 2000h

	[map all RxANSI.MAP]

	cpu 8086
	org 100h

	section TRANSIENT align=1	; Transient program part (command line parser, setup)

transient:
			; This is a simple MS-DOS 1.x detection (9 Byte).
	mov ah, 4Dh		; the only function I found that doesn't need much setup or cleanup
	stc			; but that uses MS-DOS 2.x error reporting. clears CF (Int21.30 does not)
	int 21h
	jnc .dosgood		; it is MS-DOS 2.x+ or compatible -->
	int 20h			; it isn't so simply terminate
.dosgood:
	cld
	mov dx, msg.name
	call disp_msg		; Start with message

			; The UMB link and memory allocation strategy is now always safed here
			; and restored by abort_msg as well as any other termination code. This
			; saves any code changing the strategy the effort to restore it proper.
	mov ax, 5802h
	int 21h
	xor ah, ah		; some "DOS" only return al
	mov word [ restored.umblink ], ax	; save UMB link

	mov ax, 5800h
	int 21h
	mov word [ restored.strategy ], ax	; save strategy

	mov dx, i23
	mov ax, 2523h
	int 21h			; set Int23 so that we restore link/strategy on Ctrl-C as well
	mov dx, i24
	mov ax, 2524h
	int 21h			; set Int24 so all hard errors are changed to the soft error

getfirstumb:
			; We try to get the first UMB for various uses later. Note that we
			; never use the result as an MCB address so if it's wrong we won't
			; crash or something. (We compare memory block or MCB addresses to
			; the value to check whether they're above or below it.)
	mov ax, 1261h		; PTS-DOS: Get first UMCB
	stc
	int 2Fh
	jc .determine		; not supported -->
	inc ax
	cmp ax, byte 2		; -1, 0, 1 ?
	jbe .determine		; not supported (or none) -->
	dec ax
	mov word [ firstumb ], ax	; set UMB
	jmp short .got		; got it -->

.determine:
	mov ax, 5803h
	xor bx, bx
	int 21h			; disable UMB link, leave only LMA chain
	jc .none		; that isn't supported either -->

	mov ah, 52h
	int 21h
	mov ax, word [ es:bx - 2 ]
	mov dx, ax		; first MCB
	xor bx, bx		; use offsets from bx, not addresses
.looplmb:
	mov ds, ax
	inc ax
	add ax, word [ bx + 3 ]	; next MCB's address
	cmp byte [ bx ], 'M'
	je .looplmb		; not last -->
	cmp byte [ bx ], 'Z'
	jne .none		; corrupted -->
	xchg ax, dx		; dx = what we assume to be the first UMA chain MCB
				; ax = first MCB

	push ax
	inc bx			; = 1
	mov ax, 5803h
	int 21h			; enable UMB link, include UMA chain
	pop ax
	jc .none		; so we can disable it but not enable? -->

	dec bx			; = 0
	xor cx, cx		; flag if assumed first UMCB found
.loopumb:
	cmp ax, dx
	jne .notlastlmb
	inc cx			; there it is
.notlastlmb:
	mov ds, ax
	cmp byte [ bx ], 'M'
	jne .islastumb?		; last or corrupted -->
	inc ax
	add ax, word [ bx + 3 ]
	jmp short .loopumb	; process next -->
.islastumb?:
	cmp byte [ bx ], 'Z'
	jne .none		; corrupted -->
	jcxz .none		; didn't find that UMCB -->
			; The MCB at dx which was behind the one that contained the 'Z'
			; signature when we disabled the UMB link is now a valid MCB in
			; the MCB chain after we enabled the UMB link. All previous MCBs
			; are now 'M'.
	mov word [ cs:firstumb ], dx
.none:
.got:
				; no need to restore es here since we do that below
				; no need to restore ds since we set it below

	xor bp, bp		; initial option selector

	mov ax, cs
	add ax, 10h+transient_size_p
	mov ds, ax
	mov bx, ctrl0		; ds:bx -> resident segment's ctrl0

findinstalled:
	mov ah, 0FFh		; start with multiplex number 0FFh
.loop:
	mov al, 00h		; AMIS installation check
	int 2Dh			; AMIS (or "DOS reserved" = iret if no AMIS present)
	cmp al, 0FFh
	je .try
.notfound:
	sub ah, 1		; search is backward (to find latest installed first), from 0FFh to 00h including
	jnc .loop		; try next if we didn't checked all yet -->

	jmp short scancl	;  If not found, start command line processing -->

.try:
	mov si, amissig		; ds:si -> our AMIS name strings
	mov es, dx		; es:di -> name strings of AMIS multiplexer that just answered
	mov cx, 16		; Ignore description, only compare vendor and program name
	repe cmpsb
	jne .notfound		; No match, try next
	mov ds, dx		; set ds to our resident copy's segment
	mov al, 10h		; get compatible RxANSI version
	int 2Dh
	cmp al, 0FFh		; al = 0FFh if supported
	jne .unknown		; doesn't support this question -->
	cmp cx, RXANSI_VERC	; Compatible version?
	jne .unknown
	mov al, 20h		; Get RxANSI ctrl0
	int 2Dh
	cmp al, 00h		; al = size of returned ctrl data. at least one byte if supported
	jne .known		; if size ok, known -->
.unknown:	
	or bp, _incompatresident; indicate resident copy is incompatible
.known:
	or bp, _foundresident	; indicate resident copy found

scancl:			; bp = selected options
			; ah = AMIS multiplex number of resident copy if bp&_foundresident (else 0FFh)
			; ds:bx-> ctrl0 of either resident copy or stored handler

	mov si, 81h		; PSP command line
	push cs
	pop es			; es changed while searching for resident
.loop:	cs lodsb		; ds is resident segment, so use cs
	cmp al, 9		; tab ?
	je .loop
	cmp al, 32		; blank ?
	je .loop
	jb .end			; zero, cr, lf ? -->
	cmp al, '-'
	je .switch
	cmp al, '/'
	jne .invalid		; unknown character -->
			; parse switch
.switch:
	cs lodsb		; get next character
	cmp al, 'a'
	jb .notlowercase
	cmp al, 'z'
	ja .notlowercase
	and al, ~ 20h		; Uppercase if required
.notlowercase:
			; check against known switches
	push si
	push ax
	mov si, parameters
	xchg al, ah
.loopswitch:
	cs lodsb
	xchg ax, cx
	cs lodsb
	xchg ax, cx		; cx = option selector bits
	cmp al, 0FFh
	je .invalid		; not found!
	cmp al, ah
	jne .loopswitch		; try next -->
	pop ax			; %$&" preserve ah (multiplex number) !
	pop si
	xor ch, ch

	test cl, _state		; ANSI/LESS state ?
	jz .normalbp		; no, normal option --> (sets bit in bp)

		; cx = bit to set or clear in bp
		; cx<<2 = bit to set in bp to show the cx bit was set
		; ds:si-> '+' to set, '-' to clear
	cs lodsb
	cmp	al,'-'		; OFF ?
	jne	.notminus
	call	flagset
	not	cx		; reverse bits
	and	bp,cx		; do bp&!cx
	jmp short .loop
.notminus:
	cmp	al,'+'		; ON ?
	jne	.invalid
	call	flagset
				; then fall through to do bp|cx
.normalbp:
	or bp, cx
	jmp short .loop		; ds:si-> character after switch character. just continue

.invalid:			; display help if invalid
	mov dx, msg.help
	jmp short .abort_msg

.end:
	test bp, _uninstall
	jz .dontuninstall
	test bp, _installlow|_installnew| \
		_ansiset|_lessset
	jnz .invalid		; all invalid if /u selected
	jmp uninstall

.dontuninstall:
	test bp, _foundresident	; no resident found ?
	jz .install		; yes, install -->
	test bp, _installnew	; forced to install new ?
	jz .dontinstall		; no -->

.install:
	jmp install

.dontinstall:
	test bp, _installlow
	jz .changeflags
	mov dx, msg.already	; can't install new, already resident
	jmp short .abort_msg

.changeflags:
	test bp, _incompatresident
	jz .compatible
	mov dx, msg.incompatible ; can't change flags of incompatible version
.abort_msg:
	jmp abort_msg

.compatible:
	push ax
	call changeflags
	pop ax			; AMIS multiplex number
	mov al, 20h
	int 2Dh			; clear unsupported bits

		; Display status
	mov dx, msg.stateansi	; "state ANSI: "
	call disp_msg
	test byte [bx],80h
	call onoff		; "ON (+)" or "OFF (-)" depending on ZF
	mov dx, msg.stateless	; ", LESS: "
	call disp_msg
	test byte [bx],40h
	call onoff
	mov dx, msg.crlf
	call disp_msg
	call restorestate
	mov ax, 4C00h
	int 21h

flagset:
	push cx
	rol cx, 1
	rol cx, 1
	or bp, cx
	pop cx
	retn


		; INP:	al = interrupt number
		;	ds:si-> interrupt entry
		; OUT:	CY if unhooking failed
		;	NC if unhooking successful
		; CHG:	ah, es, di
UnhookInterrupt:
			; UnhookInterruptSim (below) only checks if it's possible to unhook this interrupt.
			; This function really unhooks the interrupt if possible.
			;
			; This is to cover the situation when some of the hooked interrupts can unhook,
			; but some can't. If the uninstaller would start to unhook the interrupts and then
			; catch the interrupt that can't be unhooked the user would end with a dead TSR
			; that's uninstalled halfway. Very bad.
			;
			; "Simulating" the unhooking first and checking if all Interrupts can unhook
			; usually will not return such a state.
	call UnhookInterruptSim
	jc .ret				; bad. -->
	jz .easy
.hard:
				; "hard" case: UnhookInterruptSim has however already done the work,
				; so the hard case is here indeed easier than the easy case.
	cmpsw				; skip 10EBh or 0EA90h
	movsw				; copies our stored pointer into the other's entry
	movsw				; copies our stored segment into the other's entry
	clc
	retn
.easy:
	push ds
	push dx
	lds dx, [ si + 2 ]		; get what we stored in the entry
	mov ah, 25h			; easy case - just reset to the value stored
	int 21h				; doesn't alter CF (leaves NC from UnhookInterruptSim) or sets NC
	pop dx
	pop ds
.ret:
	retn

		; INP:	ds:si-> IISP entry
		;	al = interrupt number
		; OUT:	NC if no error (either hard or easy case),
		;	 ZR if easy case,
		;	  ds:si-> our IISP entry, containing stored interrupt
		;	 NZ if hard case,
		;	  ds:si-> our IISP entry
		;	  es:di-> IISP entry to modify
		;	CY if error (not first handler and no IISP chain to this handler)
		; CHG:	ah, es, di
UnhookInterruptSim:
	push bx
	mov ah, 35h			; get current vector
	int 21h				; es:bx-> current interrupt handler
	cmp si, bx			; our pointer ?
	jne .hard

	push ax
	push si
	mov si, ds
	mov ax, es
	 cmp si, ax			; our segment ?
	pop si
	pop ax
	jne .hard

	and ah, 00h			; NC, ZR
	pop bx
	retn
.hard:
		; INP:	ds:si-> IISP entry
		;	es:bx-> current interrupt entry
		; OUT:	CY if error
		;	NC, NZ if no error,
		;	 ds:si-> our IISP entry
		;	 es:di-> IISP entry to modify
		; CHG:	ah, es, di, (bx)
	call SearchIISPChain
	jne .harder
.found:				; found reference to our interrupt handler
	mov di, bx			; es:di-> IISP entry that references our's
	or ah, 0FFh			; NC, NZ
	pop bx
	retn

.harder:			; Desperate attempt to find IISP entry that references our's by
				; searching through the interrupts hooked by other AMIS TSRs. Note
				; that the plexer loop will find and search through the list of
				; hooked interrupts of the uninstalling TSR itself, but this causes
				; no trouble.
		; INP:	ds:si-> IISP entry
		; OUT:	CY if error
		;	NC, NZ if no error,
		;	 ds:si-> our IISP entry
		;	 es:di-> IISP entry to modify
		; CHG:	ah, es, di, (bx)
	d4bp
	push dx
	push ax				; register with interrupt number last
	xor ax, ax
.loopplex:
	mov al, 00h			; AMIS installation check
	push cx
	int 2Dh				; enquire whether there's anyone
	pop cx				;  but we don't care who it might be
	inc al
	jz .search
.nextplex:
	inc ah
	jnz .loopplex			; try next multiplexer -->
	pop ax
	pop dx
.fail:					; IISP incompatible TSR between current interrupt entry and our entry
					;  and no AMIS compatible TSR installed on top of our entry
	stc
	pop bx
	retn

		; INP:	ah = multiplex number of AMIS TSR to search through
		;	ss:sp-> interrupt number (byte), must be preserved
		; CHG:	es, di, dx, bx
.search:
	mov al, 04h
	pop bx
	push bx				; low byte is the interrupt number
	int 2Dh
	cmp al, 03h			; returned its interrupt entry ?
				; RBIL doesn't explicitly state that this interrupt entry has to
				; be IISP compatible. But I'm too lazy to look up the older AMIS,
				; and SearchIISPChain checks the interrupt entry anyway.
	je .search_dxbx
	cmp al, 04h			; returned list of hooked interrupts ?
	jne .nextplex			; no, try next multiplexer -->
	mov es, dx
	mov di, bx			; es:di-> list
	pop bx
	push bx				; bl = interrupt number
	mov al, bl
.search_intlist:		; Search the returned list for the required interrupt number.
	scasb				; our interrupt number ?
	je .search_found
	cmp byte [es:di-1], 2Dh		; was last in list ?
	je .nextplex
	scasw				; skip pointer
	jmp short .search_intlist	; try next entry -->
.search_found:
	mov bx, word [es:di]		; dx:bx-> IISP entry
.search_dxbx:
	mov es, dx			; es:bx-> (IISP) interrupt entry
				; The entry we found now is possibly behind the non-IISP entry that
				; terminated our first SearchIISPChain call (at .hard). We then
				; possibly might find our entry in this hidden part of the chain.
	call SearchIISPChain
	jne .nextplex			; didn't found our entry in the chain -->
	pop ax
	pop dx
	jmp short .found


SearchIISPChain.next:
	les bx, [es:bx +2]		; get next interrupt entry

		; INP:	ds:si-> IISP entry
		;	es:bx-> current interrupt entry
		; OUT:	NZ if reference to ds:si not found in IISP chain es:bx->
		;	ZR if reference found,
		;	 es:bx-> IISP (or uninstalled iHPFS) interrupt entry with reference
		; CHG:	es, bx
SearchIISPChain:
	call IsIISPEntry?			; that an IISP entry ?
	jnz .return				; nope -->
	cmp si, word [ es:bx + ieNext ]		; our pointer ?
	jne .next				; no, try next -->
	push ax
	mov ax, ds
	cmp ax, word [ es:bx + ieNext + 2]	; our segment ?
	pop ax
	jne .next				; no, try next -->
.return:					; yes, found
	retn


uninstall:
	d4bp
	test bp, _foundresident
	jnz .found
	mov dx, msg.notinstalled	; nothing to uninstall there
	jmp short .abort_msg

.found:
	test bp, _incompatresident
	jz .compat
	mov dx, msg.incompatible	; resident, but incompatible version
	jmp short .abort_msg

.compat:
				; ds:bx-> ctrl0 of resident copy (unused yet)
				; ah = AMIS multiplex number of resident copy
	mov al, 02h
	mov dx, cs
	mov bx, .done			; dx:bx = return address if successful
	int 2Dh				; AMIS uninstall TSR
	cmp al, 0FFh			; 0FFh successful
	je .done			; TSR has already done anything -->
	test al, al			; 00h not implemented
	jz .continue_noseg		; do it myself -->
	cmp al, 03h			; 03h safe, no resident uninstaller (still enabled). bx=segment
	je .continue_seg		; expected -->
	cmp al, 04h			; 04h safe, no resident uninstaller (now disabled). bx=segment
	je .continue_seg		; unexpected, but continue -->
		; (other values: 01h unsuccessful or internal, 02h can't uninstall yet,
		; but will do so when able, 05h not safe, 06h,07h device driver; all fail)
.fail:
	mov dx, msg.cantuninstall	; some error occured
	call disp_msg
	mov dl, '.'
	mov ah, 02h
	int 21h				; display dot to end message
	mov dx, msg.crlf
.abort_msg:
	jmp short abort_msg

.continue_noseg:
	xor bx, bx			; = 0

.continue_seg:
	push bx

	mov al, 04h
	int 2Dh				; AMIS determine hooked interrupts
	cmp al, 04h
	jne .fail		; General uninstallers should be prepared for at least 04h or 03h
				; as return code here.
	mov ds, dx			; ds:bx-> interrupt table
	push bx
	mov dx, UnhookInterruptSim
	mov cx, .simulated
	jmp short loopamisintr
.simulated:
	jc unhookerror
	pop bx				; restore ds:bx->

	mov dx, UnhookInterrupt
	mov cx, .unhooked
	jmp short loopamisintr
.unhooked:
	jc unhookerror

	push ds
	pop es				; es = segment of interrupt list
	pop ax				; stacked segment
	or ax, ax			; marker to use interrupt list segment ? (zero)
	jz .uselistsegment		;  yes -->
	mov es, ax			; else use segment returned by uninstall function

.uselistsegment:
	mov ah, 49h			; Free memory
	int 21h
				; General deinstallation code should not assume the TSR was just in
				; that single memory block. If the deinstallation code doesn't know
				; the TSR, it should search (and if it finds any, free) memory blocks
				; that have the MCB owner either that's now in es (code segment of TSR)
				; or the same MCB owner value that the TSR's code segment's MCB had.
				; (Only if the TSR's code segment was a valid memory block with MCB.)
	jc .fail			; if that causes an error, still report "fail"
					; (but interrupts are already unhooked now)

.done:
	mov dx, msg.removed
	call disp_msg			; "removed."
..@stack equ $ - (($-$$) % 2)		; (above is overwritten by stack while installing)
..@exit:				; terminate successful, re-used by some code (install too)
	call restorestate
	mov ax, 4C00h
	int 21h


unhookerror:
	mov dx, msg.cantuninstall
	call disp_msg

	 pop dx
	cmp dx, byte -1			; at least another interrupt failed ? (at least two interrupts failed ?)
	 push dx
	je .single
	mov dx, msg.interrupts		; first message for multiple interrupts
	jmp short .first
.single:
	mov dx, msg.interrupt		; first message for single interrupt
.first:
	mov bx, dx

.loop:
	call disp_msg			; msg.interrupts or msg.hcomma
	call disp_al_hex
	pop ax				; get next value from stack
	 pop dx
	cmp dx, byte -1			; at least two more ?
	 push dx
	mov dx, msg.hcomma		; message if at least two more interrupts to display
	jne .notlast			; yes -->
	mov dx, msg.hand		; only one more interrupt to display
.notlast:
	cmp ax, byte -1			; last error marker?
	jne .loop			; no, keep looping -->

	mov dx, msg.hooked		; last message for single interrupt
	cmp bx, msg.interrupt		; was first message for single interrupt ?
	je .abort			; yes -->
	mov dx, msg.hookeds		; else use last message for multiple interrupts
.abort:
	; jmp short abort_msg

abort_msg:
	call disp_msg
	call restorestate
	mov ax, 4CFFh
	int 21h


		; INP:	ds:bx -> AMIS interrupt list
		;	dx = code called for each interrupt (with ds:si-> interrupt entry)
		;	cx = code to return to
		; OUT:	CY if any interrupt failed,
		;	 ax = first failed interrupt
		; 	 ss:sp-> 0FFFFh terminated list of failed interrupts on stack
		;	NC if all interrupts successful uninstalled
		; CHG:	si, ax, bx, what called code (dx) changes
		;
		; Called code (dx) has to preserve dx, bx, cx, al.
loopamisintr:
	
	mov ax, 0FFFFh
	push ax
.loop:
	mov al, byte [ bx ]
	mov si, word [ bx+1 ]

	call dx
	jnc .noerror			; no error -->
	xor ah, ah
	push ax				; else remember number of interrupt, but continue looping

.noerror:
	add bx, byte 3
	cmp al, 2Dh
	jne .loop			; do until 2Dh done -->

	pop ax
	cmp ax, byte -1			; if it's below (CY), there were errors
	jmp cx


changeflags:
				; bp = option selector
				; ds:bx-> ctrl0
	mov ax, bp
	test ah, _ansiset>>8
	jz .noansi
	push ax
	and al, _ansistate
	and byte [bx], ~_ansistate
	or byte [bx], al
	pop ax
.noansi:
	test ah, _lessset>>8
	jz .noless
	and al, _lessstate
	and byte [bx], ~_lessstate
	or byte [bx], al
.noless:
	retn


install:		; ds = handler segment inside cs
			; bp = option selector
	d4bp
	mov ax, cs
	add ax, 10h+transient_size_p
	mov ds, ax		; ds = handler segment (could have changed with /n switch)

	mov ah, 0Fh
	int 10h			; get video mode etc, bh = active video page
	mov ah, 08h		; read character and attribute at cursor position
	int 10h
	mov byte [attrib], ah	; inherit current attribute

	mov bx, ctrl0		; ds:bx -> ctrl0
	call changeflags	; set flags from bp

	xor ax, ax
	xchg ax, word [ cs:2Ch ]; set PSP field to zero
	mov es, ax
	mov ah, 49h
	int 21h			; Free our environment
	mov sp, ..@stack	; change to stack inside now unused code

	mov ah, 4Ah
	push cs
	pop es			; Resize the segment allocated to this program
	mov bx, 10h+transient_size_p+handler_size_p	; 10h is the size of the PSP
%if _PSPRELOC
	push bx
%endif
	int 21h

%if _PSPRELOC
	xor bx, bx		; disable UMB link
	mov dx, 2		; last fit, low then high
	call setstrat
	mov dx, msg.mcbcorrupted
	jc .abort_msg
	pop bx
	mov ah, 48h
	int 21h			; allocate new memory for process
	jc .cantrelocate	; not possible, try to allocate resident -->
	mov es, ax		; newly allocated block
	xor di, di
	xor si, si
	mov cx, (10h+transient_size_p+handler_size_p)*8
	rep cs movsw		; copy process into newly allocated block
	mov dx, cs		; dx = old process

	push ax			; new cs
	mov bx, .relocated
	push bx			; new ip
	retf
.relocated:
	mov ss, ax		; change stack, too. sp is still valid!
	call setmcb
%if 1
	mov es, ax		; es = new PSP
	mov ds, dx		; ds = old PSP
	mov di, 18h
	mov cx, 20
	mov word [ es:32h ], cx
	mov word [ es:34h ], di
	mov word [ es:34h+2 ], ax	; fix the new PSP's PHT pointer
	push ax
	push di
	mov al, -1
	rep stosb		; initialize new PHT with empty entries
	pop di
	mov cx, word [ 32h ]	; cx = count of PHT entries
	cmp cx, 20		; > 20 ?
	jbe .normaltable	; no -->
	mov cx, 20		; limit to 20
.normaltable:
	lds si, [ 34h ]		; ds:si-> old PHT
	push si
	push cx
	rep movsb		; get all entries
	pop cx
	pop di
	 push ds
	 pop es			; es:di-> old PHT
	rep stosb		; fill moved entries with zero
	pop ax
	mov ds, dx		; ds = old PSP
	mov word [ 0Ah ], .terminated
	mov word [ 0Ah+2 ], ax
	mov word [ 0Eh ], i23
	mov word [ 0Eh+2 ], ax
	mov word [ 12h ], i24
	mov word [ 12h+2 ], ax	; set interrupt vectors to ours
	mov word [ 16h ], ax	; set parent PSP to the relocated one
	push word [ 2Eh ]
	pop word [ cs:2Eh ]
	mov word [ cs:2Eh+2 ], ax	; set SS:SP used by process termination
				; In order to set the correct stack address here,
				; the last Int21 call to a usual function (such as
				; Int21.48) must've been made with the same stack
				; pointer as the Int21.4C call below gets.

	xchg ax, bx		; bx = new location, dx = old location
	mov ax, 335Dh
	int 21h			; PSP relocated call-out

	mov ax, 4C00h
	int 21h			; terminate, and make the new PSP active
				; also handles freeing all memory allocated to the old PSP
				; also closes any handles >20 if PHT larger
				; also relocates Int23, Int24
				; also notifies resident software old PSP is no longer valid
.terminated:			; (ax, bx, es might be changed)
	push cs
	pop ds
%else
	mov word [ cs:34h+2 ], ax	; fix up the PHT pointer
	xchg ax, bx
	mov ah, 50h
	int 21h			; set current PSP to new
				; bx = new PSP location
				; dx = old PSP location
	mov ax, 335Dh
	int 21h			; PSP relocated call-out

	push dx
	 push cs
	 pop ds
	mov dx, i23
	mov ax, 2523h		; set Int23 to relocated position
	int 21h
	mov dx, i24
	mov ax, 2524h		; set Int24 to relocated position
	int 21h
	pop es			; former process
	mov ah, 49h
	int 21h			; free memory of former process location
%endif
	mov ah, 1Ah
	mov dx, 80h
	int 21h			; set DTA

.cantrelocate:		; We could display an error message here.
	mov bx, cs
	add bx, 10h+transient_size_p
	mov ds, bx		; ds = handler segment
%endif

	test bp, _installlow	; try UMA installation ?
	jnz .allocatelow	; nope -->
	mov bx, 1		; enable UMB link
	mov dx, 41h		; best fit, high only strategy
	call setstrat
	jc .allocatelow

	mov ah, 48h
	mov bx, handler_size_p	; required size (paragraphs)
	int 21h			; try to allocate in upper memory
	jnc .allocated		; successfully allocated -->
	cmp ax, byte 8		; "not enough memory" ?
	je .allocatelow		; yes, retry in low memory -->
.corrupted:			; else: MCBs corrupted
	mov dx, msg.mcbcorrupted
.abort_msg:
	jmp abort_msg

.allocatelow:
	xor bx, bx		; disable UMB link
	mov dx, 1		; best fit, low only strategy
	call setstrat
	mov dx, msg.mcbcorrupted
	jc .abort_msg

	mov ah, 48h
	mov bx, handler_size_p	; required size (paragraphs)
	int 21h			; try to allocate in low memory
	jnc .allocated		; successfully allocated -->
	cmp ax, byte 8		; "not enough memory" ?
	jne .corrupted		; no, assume corrupted chain -->
	mov dx, msg.nomemory	; else report no memory
	jmp short .abort_msg

.allocated:
			; ax = allocated memory block
			; ds = segment of resident (behind transient)
	mov es, ax
			; es = ax
	mov cx, (handler_size_p)*8
	xor di, di		; es:di-> start of allocated block
	xor si, si		; ds:si-> source inside this block
	rep movsw

	mov ah, 00h		; start with multiplex number 00h
.loopamis:
	mov al, 00h		; installation check
	int 2Dh
	test al, al		; still 00h if free
	jz .foundamis
	inc ah			; search from bottom to top
	jnz .loopamis		; loop if zero not yet reached -->
	mov dx, msg.noamisnumber; abort if no free AMIS multiplex number left
	jmp short .abort_msg

.foundamis:
			; ah = free AMIS multiplex number
	mov byte [ es:amisnum ], ah	; Set multiplex number

			; All error conditions must be checked above because we'll now
			; modify the MCB so that it won't be deallocated when terminating.
	mov ax, es
	call setmcb
			; ax = segment of resident copy
			; cx = size of copied handler in words
	cmp ax, word [ cs:firstumb ]	; is that high ?
	mov dx, msg.highloadedusing
	ja .highloaded		; of course -->
	mov dx, msg.loadedusing	; no, other message

.highloaded:
	call disp_msg		; "[high ]loaded using "
	push ax			; allocated block
	mov ax, handler_size_p*16 +16	; +16 is for the MCB (most MEMs count it)
	call disp_ax_dec	; "<num>"
	mov dx, msg.byte
	call disp_msg		; " bytes."

	pop ds			; ds = allocated block
	mov bx, amisintr	; ds:bx -> AMIS interrupt list
	mov dx, HookInterrupt
	mov cx, ..@exit
	jmp loopamisintr	; hook all interrupts in list, return to ..@exit

		; INP:	ds:si-> IISP entry
		;	al = interrupt number
		; OUT:	NC
		; CHG:	ah, es
HookInterrupt:
	push bx
	push dx
	mov ah, 35h
	int 21h					; get current handler
	test byte [ si + ieEOI ], 80h		; should be last handler ?
	jz .notlast
	call IsChainingIISPEntry?		; is the first handler a chaining IISP ?
	jne .notlast				; no, can't hook below it -->
	db __NOP_DATA2				; (skip 2 pop ax)
.loop:
	pop ax
	pop ax
	push es
	push bx
	les bx, [ es:bx + ieNext ]		; load next entry
	call IsChainingIISPEntry?		; this one also a chaining IISP ?
	je .loop
	mov word [ si + ieNext ], bx
	mov word [ si + ieNext + 2 ], es	; store address of next entry (non-IISP or non-chaining)
	pop bx
	pop es
	mov word [ es:bx + ieNext ], si
	mov word [ es:bx + ieNext + 2 ], ds	; store our handler's address in this chaining IISP entry
	jmp short .return

.notlast:
	mov word [ si + ieNext ], bx
	mov word [ si + ieNext + 2 ], es	; store
	mov ah, 25h
	mov dx, si				; ds:dx-> interrupt entry
	int 21h					; set new handler
.return:
	pop dx
	pop bx
	clc
	retn


		; INP:	es:bx-> interrupt entry
		; OUT:	NZ if non-IISP entry,
		;	 or IISP entry that doesn't chain
		;	ZR if IISP entry that chains
IsChainingIISPEntry?:
	test byte [ es:bx + ieEOI ], 80h	; this one a non-chaining handler ? (or non-IISP)
	jnz IsIISPEntry?.return			; yes -->
						; otherwise fall through to check if really an IISP entry

		; INP:	es:bx-> interrupt entry
		; OUT:	NZ if non-IISP entry
		;	ZR if IISP entry
IsIISPEntry?:
	cmp word [ es:bx + ieSignature ], "KB"	; "KB"/424Bh ? ("BK" in MASM)
	jne .return
	cmp word [ es:bx + ieEntry ], 0EA90h	; nop\jmp far imm16:imm16 ?
	je .return				; unused IISP entry (created by iHPFS) -->
	cmp word [ es:bx + ieEntry ], 10EBh	; jmp short $+12h ?
	jne .return
	cmp byte [ es:bx + ieJmphwreset ], 0EBh	; jmp short ... ?
.return:
	retn


		; INP:	ax = memory block
		; OUT:	ds = MCB, name and owner set
setmcb:
	dec ax
	mov ds, ax		; ds = MCB of allocated block
	inc ax			; ax = allocated block!
	mov word [ 8 ], "RX"
	mov word [ 8+2 ], "AN"
	mov word [ 8+4 ], "SI"
	and word [ 8+6 ],byte 0	; Force MCB string to "RXANSI",0
	mov word [ 1 ], ax	; Set owner to itself
	retn


			; The DOS memory allocation strategy classes are actually "Low then high",
			; "High only" and "High then low". RBIL tells trash regarding that. We've
			; to disable the UMB link to insure only low memory is allocated. Normally,
			; the UMB link is already disabled when the program starts. But some
			; programs, like CMDEDIT 3.21, are buggy and leave the UMB link enabled
			; after they used UMBs.
		; INP:	bx = UMB link status to use
		;	dx = memory allocation strategy to use
		; OUT:	CF/ax error status
setstrat:
	mov ax, 5803h
	int 21h			; set UMB link
	jnc .linkdone		; no error
	cmp ax, byte 1		; "invalid function" ?
	stc
	jne .return		; no, actual error -->
	test bx, bx		; wanted to disable anyway ?
	stc
	jnz .return		; no, regard as error -->
.linkdone:
	mov ax, 5801h
	mov bx, dx
	int 21h			; set strategy
.return:
	retn


i24:
	mov al, 3			; always return fail, to handle the error as soft one
	iret

i23:
	mov word [ cs:$ ], (__JMP_REL8|__REL8__(.return-1)<<8)	; don't reenter
	call restorestate
.return:
	stc				; always abort program (what default DOS handler also does)
	retf

		; Restore modified DOS data
		;
		; CHG:	-
		; USE:	Int21
restorestate:
	push ax
	push bx
	mov bx, word [ cs:restored.strategy ]
	mov ax, 5801h
	int 21h				; restore strategy

	mov bx, word [ cs:restored.umblink ]
	mov ax, 5803h
	int 21h				; restore UMB link
	pop bx
	pop ax
	retn


onoff:
	mov dx, msg.off
	jz .off
	mov dx, msg.on
.off:				; fall through and display either
disp_msg:
	 push ds
	 push ax
	push cs
	pop ds
	mov ah, 09h
	int 21h
	 pop ax
	 pop ds
	retn


disp_al_hex:
		push cx
		mov cl, 4
		ror al, cl
		call .nibble
		ror al, cl
		pop cx
.nibble:
		push ax
		and al, 0Fh
		add al, '0'
		cmp al, '9'
		jbe .isdigit
		add al, 'A'-('9'+1)
.isdigit:
		xchg dl, al
		mov ah, 02h
		int 21h
		pop ax
		retn

	; Following call: Display number in ax decimal
	; all registers preserved except dx
disp_ax_dec:			; ax (no leading zeros)
	; In: number in ax
	; Out: displayed
		push bx
		xor bx, bx
.pushax:
		push ax
.pushend:
		or bl, bl
		jz .nobl
		sub bl, 5
		neg bl
.nobl:
		push cx
		mov cx, 10000
		call .divide_out
		mov cx, 1000
		call .divide_out
		mov cx, 100
		call .divide_out
		mov cl, 10
		call .divide_out
							; (Divisor 1 is useless)
		add al, '0'
		xchg dl, al
		mov ah, 02h
		int 21h
		pop cx
		pop ax
		pop bx					; Caller's register
		retn

.divide_out:
	; In: ax = number
	;     cx = divisor
	; Out: ax = remainder of operation
	;      result displayed
		push dx
		xor dx, dx
		div cx				; 0:ax / cx
		push dx				; remainder
		dec bl
		jnz .nobl2
		or bh, 1
.nobl2:
		or bh, al
		jz .leadingzero
		add al, '0'
		xchg dl, al
		mov ah, 02h
		int 21h				; display result
 .leadingzero:
		pop ax				; remainder
		pop dx
		retn

msg:
			; The version string is international
.name:		db "RxANSI ",RXANSI_VERP_STR,"beta4: ",36
%strcat _LANGFILE "RxANSIT.",_LANG
	%include _LANGFILE

firstumb:	dw 9FFFh			; guess of where UMBs start (A000h+)
restored:
.umblink:	dw 0
.strategy:	dw 0

parameters:			; Valid switches on the command line, followed
				; by an option selector byte with one bit set.
	db "A", _ansistate
	db "L", _lessstate
	db "J", _installlow
	db "N", _installnew
	db "U", _uninstall
	db 0FFh			; table end marker
			; The help switches are not in this table (and neither in the option
			; selector) because on invalid switches, we'll show the help anyway.

	align 16
	endarea transient


	section RESIDENT vstart=0 align=16
				; The handler is later copied into a seperate memory block.
				; The vstart= value tells NASM to use all addresses as if this
				; location is actually at vstart. (In this case, like if it's
				; at the start of the segment, not at 100h+start_size.)


%include "RXANSI.ASM"

%if _UPXPAD && (transient_size_p*16 + handler_size) < 3062
		times 3062 -(transient_size_p*16 + handler_size) db 90h
%endif